/*
 * This file is generated by gdbus-codegen, do not modify it.
 *
 * The license of this code is the same as for the D-Bus interface description
 * it was derived from. Note that it links to GLib, so must comply with the
 * LGPL linking clauses.
 */

#ifndef __IPC_PROFILER_H__
#define __IPC_PROFILER_H__

#include <gio/gio.h>

G_BEGIN_DECLS


/* ------------------------------------------------------------------------ */
/* Declarations for org.gnome.Sysprof3.Profiler */

#define IPC_TYPE_PROFILER (ipc_profiler_get_type ())
#define IPC_PROFILER(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), IPC_TYPE_PROFILER, IpcProfiler))
#define IPC_IS_PROFILER(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), IPC_TYPE_PROFILER))
#define IPC_PROFILER_GET_IFACE(o) (G_TYPE_INSTANCE_GET_INTERFACE ((o), IPC_TYPE_PROFILER, IpcProfilerIface))

struct _IpcProfiler;
typedef struct _IpcProfiler IpcProfiler;
typedef struct _IpcProfilerIface IpcProfilerIface;

struct _IpcProfilerIface
{
  GTypeInterface parent_iface;


  gboolean (*handle_start) (
    IpcProfiler *object,
    GDBusMethodInvocation *invocation,
    GUnixFDList *fd_list,
    GVariant *arg_options,
    GVariant *arg_fd);

  gboolean (*handle_stop) (
    IpcProfiler *object,
    GDBusMethodInvocation *invocation);

  GVariant * (*get_capabilities) (IpcProfiler *object);

};

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (IpcProfiler, g_object_unref)
#endif

GType ipc_profiler_get_type (void) G_GNUC_CONST;

GDBusInterfaceInfo *ipc_profiler_interface_info (void);
guint ipc_profiler_override_properties (GObjectClass *klass, guint property_id_begin);


/* D-Bus method call completion functions: */
void ipc_profiler_complete_start (
    IpcProfiler *object,
    GDBusMethodInvocation *invocation,
    GUnixFDList *fd_list);

void ipc_profiler_complete_stop (
    IpcProfiler *object,
    GDBusMethodInvocation *invocation);



/* D-Bus method calls: */
void ipc_profiler_call_start (
    IpcProfiler *proxy,
    GVariant *arg_options,
    GVariant *arg_fd,
    GUnixFDList *fd_list,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean ipc_profiler_call_start_finish (
    IpcProfiler *proxy,
    GUnixFDList **out_fd_list,
    GAsyncResult *res,
    GError **error);

gboolean ipc_profiler_call_start_sync (
    IpcProfiler *proxy,
    GVariant *arg_options,
    GVariant *arg_fd,
    GUnixFDList  *fd_list,
    GUnixFDList **out_fd_list,
    GCancellable *cancellable,
    GError **error);

void ipc_profiler_call_stop (
    IpcProfiler *proxy,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean ipc_profiler_call_stop_finish (
    IpcProfiler *proxy,
    GAsyncResult *res,
    GError **error);

gboolean ipc_profiler_call_stop_sync (
    IpcProfiler *proxy,
    GCancellable *cancellable,
    GError **error);



/* D-Bus property accessors: */
GVariant *ipc_profiler_get_capabilities (IpcProfiler *object);
GVariant *ipc_profiler_dup_capabilities (IpcProfiler *object);
void ipc_profiler_set_capabilities (IpcProfiler *object, GVariant *value);


/* ---- */

#define IPC_TYPE_PROFILER_PROXY (ipc_profiler_proxy_get_type ())
#define IPC_PROFILER_PROXY(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), IPC_TYPE_PROFILER_PROXY, IpcProfilerProxy))
#define IPC_PROFILER_PROXY_CLASS(k) (G_TYPE_CHECK_CLASS_CAST ((k), IPC_TYPE_PROFILER_PROXY, IpcProfilerProxyClass))
#define IPC_PROFILER_PROXY_GET_CLASS(o) (G_TYPE_INSTANCE_GET_CLASS ((o), IPC_TYPE_PROFILER_PROXY, IpcProfilerProxyClass))
#define IPC_IS_PROFILER_PROXY(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), IPC_TYPE_PROFILER_PROXY))
#define IPC_IS_PROFILER_PROXY_CLASS(k) (G_TYPE_CHECK_CLASS_TYPE ((k), IPC_TYPE_PROFILER_PROXY))

typedef struct _IpcProfilerProxy IpcProfilerProxy;
typedef struct _IpcProfilerProxyClass IpcProfilerProxyClass;
typedef struct _IpcProfilerProxyPrivate IpcProfilerProxyPrivate;

struct _IpcProfilerProxy
{
  /*< private >*/
  GDBusProxy parent_instance;
  IpcProfilerProxyPrivate *priv;
};

struct _IpcProfilerProxyClass
{
  GDBusProxyClass parent_class;
};

GType ipc_profiler_proxy_get_type (void) G_GNUC_CONST;

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (IpcProfilerProxy, g_object_unref)
#endif

void ipc_profiler_proxy_new (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data);
IpcProfiler *ipc_profiler_proxy_new_finish (
    GAsyncResult        *res,
    GError             **error);
IpcProfiler *ipc_profiler_proxy_new_sync (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error);

void ipc_profiler_proxy_new_for_bus (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data);
IpcProfiler *ipc_profiler_proxy_new_for_bus_finish (
    GAsyncResult        *res,
    GError             **error);
IpcProfiler *ipc_profiler_proxy_new_for_bus_sync (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error);


/* ---- */

#define IPC_TYPE_PROFILER_SKELETON (ipc_profiler_skeleton_get_type ())
#define IPC_PROFILER_SKELETON(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), IPC_TYPE_PROFILER_SKELETON, IpcProfilerSkeleton))
#define IPC_PROFILER_SKELETON_CLASS(k) (G_TYPE_CHECK_CLASS_CAST ((k), IPC_TYPE_PROFILER_SKELETON, IpcProfilerSkeletonClass))
#define IPC_PROFILER_SKELETON_GET_CLASS(o) (G_TYPE_INSTANCE_GET_CLASS ((o), IPC_TYPE_PROFILER_SKELETON, IpcProfilerSkeletonClass))
#define IPC_IS_PROFILER_SKELETON(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), IPC_TYPE_PROFILER_SKELETON))
#define IPC_IS_PROFILER_SKELETON_CLASS(k) (G_TYPE_CHECK_CLASS_TYPE ((k), IPC_TYPE_PROFILER_SKELETON))

typedef struct _IpcProfilerSkeleton IpcProfilerSkeleton;
typedef struct _IpcProfilerSkeletonClass IpcProfilerSkeletonClass;
typedef struct _IpcProfilerSkeletonPrivate IpcProfilerSkeletonPrivate;

struct _IpcProfilerSkeleton
{
  /*< private >*/
  GDBusInterfaceSkeleton parent_instance;
  IpcProfilerSkeletonPrivate *priv;
};

struct _IpcProfilerSkeletonClass
{
  GDBusInterfaceSkeletonClass parent_class;
};

GType ipc_profiler_skeleton_get_type (void) G_GNUC_CONST;

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (IpcProfilerSkeleton, g_object_unref)
#endif

IpcProfiler *ipc_profiler_skeleton_new (void);


G_END_DECLS

#endif /* __IPC_PROFILER_H__ */
